(*************************************************************
 *                Auto-assemblage de graphes                 *
 *************************************************************
   Auteur  : Fabien TARISSAN
   Fichier : eval.ml
   Date    : 09/06/04
 *************************************************************
   Description : Evaluation functions for Robot
 *************************************************************)


open Param;;

open Agent;;
open Assembly;;
open System;;

(* GENERATION DE NOMBRES *)
(*************************)

(* pour gnrer des nombres alatoires *)
Random.self_init ();;


(* FONCTIONS DE BASES *)
(**********************)

(* Pour extraire le nime lment de la liste l *)
let rec extract_rec n acc l =
  if (n=0)
  then (List.hd l, acc, List.tl l )
  else extract_rec (n-1) ((List.hd l)::acc) (List.tl l)
;;

(* renvoie le n1ime, le n2ime lment de l *)
(* ainsi que le reste de la liste *)
let extract n1 n2 l =
  let (a1,acc,l1) = extract_rec n1 [] l in
  let (a2,acc2,l2) = extract_rec n2 acc l1 in
    (a1,a2,acc2@l2)
;;

(* rcupre l'agent possdant une communication 'com' *)
(* parmi les agents en tat passif                    *)
(* renvoie cet agent, la liste de ses voisins dans le *)
(* spanning tree (moins la comm concerne) et la      *)
(* nouvelle liste des agents passifs                  *)

let recup_com com l =
  (* rcupre la com *)
  let rec recup_name acc s =
    match s with
	[] -> None
      | hd::tl -> 
	  if (hd=com)
	  then Some(acc@tl)
	  else recup_name (hd::acc) tl
  in
    (* rcupre l'agent *)
  let rec recup_rec acc l =
    match l with
	[] -> None
      | hd::tl ->
	  let res = recup_name [] hd.span in
	    match res with
		None -> recup_rec (hd::acc) tl
	      | Some(s') -> Some(hd,s',acc@tl)
  in
    recup_rec [] l
;;

(* choisit une com au hasard *)
let choose_one_com l =
  let n = Random.int (List.length l) in
    List.nth l n
;;


(* rcupre un agent  qui on passer le token *)
let recup_by_id_and_com id com l =
  let rec recup_rec acc l =
    match l with
	[] -> None
      | hd::tl ->
	  if (hd.ident = id & List.mem com (hd.cycle@hd.span))
	  then Some(hd, acc@tl)
	  else recup_rec (hd::acc) tl
  in
    recup_rec [] l
;;

(* Rcupre un agent en l'extrayant *)
(* il faut une liste non vide        *)
let choose_one l=
  let n = Random.int (List.length l) in
  let (a,acc,l') = extract_rec n [] l in
    (a,acc@l')
;;


(* Rcupre un agent *)
(* il faut une liste non vide *)
let choose_one_sans_chgt l =
  let n = Random.int (List.length l) in
  List.nth l n
;;


(* Rcupre deux agents *)
(* il faut que la liste contienne au moins deux agents *)
let choose_two l =
  let long = List.length l in
  let n1 = Random.int long in
    if n1 = long-1 
    then 
      let n2 = Random.int n1 in 
	extract n2 (n1-n2-1) l
    else
      let n2 = Random.int (long-n1-1) in 
	extract n1 n2 l
;;

(* Calcule la distance entre deux agents *)
(* renvoie type float *)
let distance a1 a2 =
  let (vect_x,vect_y) = (a2.pos_x -. a1.pos_x , a2.pos_y -. a1.pos_y) in
    sqrt (vect_x*.vect_x +. vect_y*.vect_y)
;;

(* TYPE DES EXCEPTIONS *)
(***********************)

exception EvalError of string;;



(* UNARY JOIN *)
(**************)

(* Rcupre tous les agents de passifs pouvant se lier avec g *)
let recup_all_self a1 sys =
  let rec recup ok l =
    match l with
	[] -> ok
      | a2::tl ->
	  if a2.ident != a1.ident || distance a1 a2 > dist_link
	  then recup ok tl
	  else (
	    let (r1,r2) = 
	      if a1.role < a2.role
	      then (a1.role, a2.role)
	      else (a2.role, a1.role)
	    in
	      match self_join (recup_image a1) r1 r2 sys.assembly with
		  None -> recup ok tl
		| Some(g') -> recup ((a2,g')::ok) tl
	  )
  in
    recup [] sys.passive
;;


(* tente de crer une connexion cyclique entre deux agents *)
(* renvoie les nouveaux tats des agents                   *)
let un_join a1 sys =
  let ok = recup_all_self a1 sys in
    if ok != []
    then
      let (a2,((g,s),b)) = choose_one_sans_chgt ok in
      let com = create_com() in
	a1.cycle <- com::a1.cycle;
	a2.cycle <- com::a2.cycle;
	a1.ident <- create_com();
	a1.role <- s.(a1.role);
	a1.state <- Update(a1.span, s, b, Some(g));
	sys.update <- a1::sys.update
    else sys.active <- a1::sys.active
;;



(* BINARY JOIN *)
(***************)

(* Rcupre tous les agents de 'liste' *)
(* pouvant se lier  l'agent concern *)
let recup_all_binary a1 sys =
  let rec partition ok non l =
    match l with
	[] -> (ok,non)
      | a2::tl ->
	  if distance a1 a2 <= dist_link then
	    let (g1,r1,g2,r2,b) =
	      if (recup_image a1) < (recup_image a2) 
		|| ((recup_image a1) = (recup_image a2) && a1.role <= a2.role)
	      then ((recup_image a1), a1.role,  (recup_image a2), a2.role, true)
	      else ((recup_image a2), a2.role,  (recup_image a1), a1.role, false)
	    in
	      match binary_join b g1 r1 g2 r2 sys.assembly with
		  None -> partition ok (a2::non) tl
		| Some(g') -> partition ( (a2,g')::ok) non tl
	  else partition ok (a2::non) tl
  in
    partition [] [] sys.active
;;


(* tente de crer une connexion entre deux agents *)
(* renvoie les nouveaux tats des agents          *)
(* le premier tant celui rest en mode actif     *)
let bin_join a1 sys =
  let (ok,non) = recup_all_binary a1 sys in
    sys.active <- non;
    if ok != []
    then 
      let ((a2,((g,(s1,s2)),b)),autre) = choose_one ok in
      let (l,_) = List.split autre in
	sys.active <- (l@sys.active);
	a1.ident <- create_com ();
	a2.ident <- a1.ident;
	a1.state <- Update(a1.span, s1, b, Some(g));
	a2.state <- Update(a2.span,s2, b, None);
	a1.role <- s1.(a1.role);
	a2.role <- s2.(a2.role);
	let com = create_com() in
	  a1.span <- com::a1.span;
	  a2.span <- com::a2.span;
	  sys.update <- a1::(a2::sys.update)
    else sys.active <- a1::sys.active
;;



(* UPDATE *)
(**********)

(* tente de mettre  jour un agent *)
(* renvoie le nouvel tat de l'agent ains *)
(* que la liste des agents en mode passif *)
let update a sys =
  match a.state with
      Update(l,s,b,e) -> (
	match l with
	    [] ->(
	      if b 
	      then (a.state <- Final;
		    sys.final <- a::sys.final )
	      else
		match e with
		    Some(g) -> 
		      a.state <- Active(g);
		      sys.active <- a::sys.active
		  | None ->
		      a.state <- Passive;
		      sys.passive <- a::sys.passive
	    )
	  | hd::tl -> 
	      let res = recup_com hd sys.passive in
		match res with
		    None -> sys.update <- a::sys.update
		  | Some(a2,s',pass') ->
		      a.state <- Update(tl,s,b,e);
		      a2.ident <- a.ident;
		      a2.role <- s.(a2.role);
		      a2.state <- Update(s',s,b,None);
		      sys.update <- a::(a2::sys.update);
		      sys.passive <- pass'
      )
    | _ -> raise (EvalError "ANORMAL SITUATION : not in an update phase")
;;



(* SWITCH *)
(**********)

(* Fait passer l'image de la composante connexe entre *)
(* deux agents, le premier tant celui en mode actif *)
let switch a sys =
  let g = recup_image a in
    if Random.int ((snd g)*(snd g) + 3) = (snd g)*(snd g) + 2
    then (
      a.span <- [];
      a.cycle <- [];
      a.state <- Alarme(Sys.time());
      sys.break <- a::sys.break
    )
    else
      match a.span with
	  [] -> sys.active <- a::sys.active
	| _ -> 
	    let com = choose_one_com (a.cycle@a.span) in
	    let res = recup_by_id_and_com a.ident com sys.passive in
	      match res with
		  None -> sys.active <- a::sys.active
		| Some(a2, pass') ->
		    a2.state <- Active(g);
		    a.state <- Passive;
		    sys.active <- a2::sys.active;
		    sys.passive <- a::pass'
;;



(* BREAK *)
(**********)

(* Propagation de l'alarme *)
let break a sys =
  let t = recup_alarme a in
(*    if (t> skip_period) *)
    if (Sys.time() -. t > skip_period)
    then
      let (g,r) = sys.assembly.init in
	a.role <- r;
	a.state <- Active(g);
	a.ident <- create_com();
	sys.active <- a::sys.active
    else 
      begin
	a.state <- Alarme(t);
	sys.break <- a::sys.break
      end
;;


(* EVALUATION *)
(**************)

(* Evaluation *)
let eval sys choice =
  match choice with
      BinJoin ->  
	if sys.active != []
	then
	  let (a1,act') = choose_one sys.active in
	    sys.active <- act';
	    bin_join a1 sys
	else ()
	  
    | UnJoin ->
	if sys.active != []
	then 
	  let (a,act') = choose_one sys.active in
	    sys.active <- act';
	    un_join a sys
	else ()
	  
    | Upd ->
	if sys.update != []
	then
	  let (a,upd') = choose_one sys.update in
	    sys.update <- upd';
	    update a sys
	else ()
	  
    | Switch ->
	if (sys.active != [])
	then
	  let (a,act')  = choose_one sys.active in
	    sys.active <- act';
	    switch a sys
	else ()

    | Break ->
	if (sys.break != [])
	then
	  let (a,bre')  = choose_one sys.break in
	    sys.break <- bre';
	    break a sys
	else ()
;;

